import dataiku
from datetime import datetime
import dash
from dash import html
from dash import dcc
import dash_bootstrap_components as dbc
from dash import dash_table
import dash_daq as daq
from dash.dependencies import Input, Output, State
import dash_interactive_graphviz
from process_mining.graph_creation import create_graph
from process_mining.process_mining import mine_process
from process_mining.filtering import *
from webapp.filters import contained_intersecting, filter_title, start_step, end_step, case_performance_slider, attribute_filter, filter_element, single_datetime_filter_picker
from webapp.selected_elements import selected_node_window, selected_edge_window
from webapp.styles import *
from data_management import get_managed_folder_id_with_folder_name, write_pickle_in_dss_folder, read_pickle_from_dss_folder
from formatting import human_format, formatted_human_readable, start_week
import pm4py
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
import pickle
import ast
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import numpy as np
from random import sample
from dash.exceptions import PreventUpdate

palette = '#D5D9D9', '#3075AE', '#ff7e0b'

old_current_process = None
current_process = None
final_process = None
individual_data = None

app.config.external_stylesheets = [dbc.themes.BOOTSTRAP, "https://use.fontawesome.com/releases/v5.13.0/css/all.css"]

timeframe = dataiku.Dataset('timeframe')
timeframe_df = timeframe.get_dataframe()

workflow = dataiku.Dataset("workflow_clean")
workflow_df = workflow.get_dataframe()

total_cases = len(workflow_df['case'].unique())

start_end = dataiku.Dataset("start_end")
start_end_df = start_end.get_dataframe()

case_performance = dataiku.Dataset("workflow_case_performance")
case_performance_df = case_performance.get_dataframe()

activities_count = dataiku.Dataset("activity_count")
activities_count_df = activities_count.get_dataframe()

activity_count = dataiku.Dataset("workflow_by_activity")
activity_count_df = activity_count.get_dataframe()

transitions_frequency_share = dataiku.Dataset("transitions_frequency_share_prepared")
transitions_df = transitions_frequency_share.get_dataframe()

client = dataiku.api_client()
project = client.get_project(dataiku.default_project_key())
variables = project.get_variables()

events_filters = variables['standard']['filtered_events']
min_time_filter = variables['standard']['min_time_filter']
self_transition = variables['standard']['self_transition']

palette = ['#BDD8ED', '#3075AE', '#4F934F']

case_performance_scale = np.percentile(case_performance_df['case_performance'], range(101))

remove_self_transition = daq.BooleanSwitch(id='self-transition',
                                           on=self_transition,
                                           persistence=False,
                                           style={'float': 'left'})

process_input_name = dcc.Input(id='saved-process-name',
                               type='text',
                               placeholder='Process Name',
                               style={'height': '36px', 'width': '100%'})

attributes_columns = [col for col in workflow_df.columns if col not in ['case', 'activity', 'timestamp']]
attribute_scales = {col: np.nanpercentile(workflow_df[col], range(101)) for col in attributes_columns if is_numeric_dtype(workflow_df[col])}

def variants_filter_object(default=10, max=100, tab='process-discovery'):
    if tab == 'process-discovery':
        display = 'block'
    else:
        display = 'None'
    variants_input = dcc.Input(
            id="variants-filter",
            type="number",
            value=default,
            min=1,
            max=max, 
            size=2,
            style={'float': 'left','margin-left': '4px', 'margin-right': '4px', 'width': '36px'})
    variants_filter = html.Div([html.P('Display the top ', style={'float': 'left','margin': 'auto', 'margin-top': '4px'}), 
                                variants_input, 
                                html.P(' variants (unique trace)', 
                                       style={'float': 'left', 'margin': 'auto', 'margin-top': '4px'})], 
                                style={'display': display})
    return variants_filter

def time_before_filter_object(default=0):
    time_before_input = dcc.Input(
            id="time-before-activity-filter",
            type="number",
            value=default,
            min=0, 
            size=2,
            style={'float': 'left','margin-left': '4px', 'margin-right': '4px', 'width': '36px'})
    time_before_filter = html.Div([html.P('Remove activities occuring less than ', style={'float': 'left','margin': 'auto', 'margin-top': '4px'}), 
                                time_before_input, 
                                html.P(' seconds than the previous one', style={'float': 'left','margin': 'auto', 'margin-top': '4px'})])
    return time_before_filter

def attribute_filters(filters=None):
    if filters is None:
        return [attribute_filter(col, workflow_df, min_value=human_format(attribute_scales[col][0]), max_value=human_format(attribute_scales[col][-1])) if is_numeric_dtype(workflow_df[col]) else attribute_filter(col, workflow_df) for col in attributes_columns]
    else:
        return [attribute_filter(col, workflow_df, value=filters[col], min_value=human_format(attribute_scales[col][0]), max_value=human_format(attribute_scales[col][-1])) if is_numeric_dtype(workflow_df[col]) else attribute_filter(col, workflow_df, value=filters[col]) for col in attributes_columns]

def legend_object(min_value, max_value):
    return [dbc.Row([
                       dbc.Col(html.P('Number of activity occurence'), 
                               style={'padding': 0, 'float': 'right'}, md=4),
     dbc.Col(dbc.Row([dbc.Col(html.P(str(min_value), style={'float': 'right',
                                'padding': 0}), style={'padding': 0}, md=3),
     dbc.Col(html.Div(style={'height': '50%',
                                   'width': '100%',
                                   'margin': 'auto',
                                   'background': 'linear-gradient(to right, #4192a9, #DFB717, #BCBB46, #e14a08)',
                                   'padding': 0}), md=6),
     dbc.Col(html.P(str(max_value), style={'padding': 0}), 
             style={'padding': 0}, md=3)]), style={'float': 'left'}, md=8)])]
    
workflow_df['datetime'] = workflow_df['timestamp'].dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
workflow_df['epoch_timestamp'] = pd.to_datetime(workflow_df['datetime']).astype(int) / 10**9
workflow_df = workflow_df.sort_values(['case', 'epoch_timestamp'])
workflow_df['diff_time'] = workflow_df.groupby(['case'])['epoch_timestamp'].diff().fillna(10000)
workflow_df = workflow_df.drop(['datetime', 'epoch_timestamp'], axis=1)
workflow_df['previous_activity'] = workflow_df.groupby(['case'])['activity'].shift(1)

@app.callback(Output('nb-cases-info', 'children'),
              Output('graph-legend', 'children'),
              Output('graph-workflow', 'children'),
              Output('unique-nodes', 'data'),
              Output('old-variants-filter', 'data'),
              State('graph-workflow', 'children'),
              State('unique-nodes', 'data'),
              State('old-variants-filter', 'data'),
              Input('variants-filter', 'value'),
              Input('frequency-time', 'on'),
              Input('tabs', 'value'))
def render_content(old_graph, unique_nodes, old_nb_variants, 
                   nb_variants, frequency_time, tab):
    global old_current_process
    global final_process
    ctx = dash.callback_context
    if old_nb_variants==nb_variants and old_current_process==current_process and ctx.triggered[0]['prop_id'] == 'variants-filter.value' and old_graph is not None:
        raise PreventUpdate
    if current_process is not None:
        current_process_parsed = pd.read_json(current_process)
        filtered_workflow = filter_workflow(current_process_parsed, nb_variants)
        filtered_cases = len(current_process_parsed['case'].unique())
        displayed_cases = len(filtered_workflow['case'].unique())
        
        if filtered_cases == 0:
            return [html.P('No cases filtered, try releasing the constraints', style={'font-size': '12px', 'margin': '1px'})], legend_object(1, 1), dbc.Col([None], style={'width': '600px'}), None, 0
        
        if frequency_time == False:
            workflow_mined, start_end, activities = mine_process(filtered_workflow, 'case', 'activity', 'timestamp', transition='frequency')
        else:
            workflow_mined, start_end, activities = mine_process(filtered_workflow, 'case', 'activity', 'timestamp', transition='performance')
        graph, unique_nodes, max_count = create_graph(workflow_mined, start_end, activities, transition='time' if frequency_time else 'frequency')
        graph_object = dash_interactive_graphviz.DashInteractiveGraphviz(
            id="filtered-workflow",
            dot_source=graph,
            style={'width': '67%', 'height': '120%',
                  'cursor': 'pointer'}
        )
        final_process = filtered_workflow.to_json()
        old_current_process = current_process
        return [html.P(human_format(displayed_cases) + ' cases displayed on ' + human_format(total_cases) + ' total', style={'font-size': '12px', 'margin': '1px'})], legend_object(1, max_count), dbc.Col([graph_object], style={'width': '600px'}), unique_nodes, nb_variants
    else:
        raise PreventUpdate

def apply_filters(workflow_df, start_date, end_date, contained,
                 start_steps, end_steps, case_performance,
                 excluded_activities, time_before_activity, self_transition,
                 *attributes):
    filtered_workflow = filter_activities(workflow_df, excluded_activities)
    filtered_workflow = filter_time_before(filtered_workflow, time_before_activity)
    if self_transition:
        filtered_workflow = filter_self_transition(filtered_workflow)
    if 'Z' in start_date:
        start_date = datetime.strptime(start_date, '%Y-%m-%dT%H:%M:%S.%fZ')
    elif '+00:00' in start_date:
        try:
            start_date = datetime.strptime(start_date, '%Y-%m-%dT%H:%M:%S+00:00')
        except:
            start_date = datetime.strptime(start_date, '%Y-%m-%dT%H:%M:%S.%f+00:00')
    else:
        start_date = datetime.strptime(start_date, '%Y-%m-%dT%H:%M:%S')
    if 'Z' in end_date:
        end_date = datetime.strptime(end_date, '%Y-%m-%dT%H:%M:%S.%fZ')
    elif '+00:00' in end_date:
        try:
            end_date = datetime.strptime(end_date, '%Y-%m-%dT%H:%M:%S+00:00')
        except:
            end_date = datetime.strptime(end_date, '%Y-%m-%dT%H:%M:%S.%f+00:00')
    else:
        end_date = datetime.strptime(end_date, '%Y-%m-%dT%H:%M:%S')
    filtered_workflow = filter_timestamp(filtered_workflow, start_date, end_date, contained)
    if start_steps:
        filtered_workflow = filter_start_step(filtered_workflow, start_steps)
    if end_steps:
        filtered_workflow = filter_end_step(filtered_workflow, end_steps)
    filtered_workflow = filter_case_performance(filtered_workflow, case_performance[0], case_performance[1], case_performance_scale)
    for i, attribute in enumerate(attributes):
        if attribute is not None:
            if len(attribute)>0:
                filtered_workflow = filter_attributes(filtered_workflow, attributes_columns[i], attribute, attribute_scales)
    return filtered_workflow

@app.callback(Output('filters', 'data'),
              Output('variants-container', 'children'),
              State('old-variants-filter', 'data'),
              Input('tabs', 'value'),
              Input('start-date-filter', 'value'),
              Input('end-date-filter', 'value'),
              Input('contained-intersecting', 'on'),
              Input('start-step', 'value'),
              Input('end-step', 'value'),
              Input('case-performance-filter', 'value'),
              Input('excluded-activities-dropdown', 'value'),
              Input('time-before-activity-filter', 'value'),
              Input('self-transition', 'on'),
              [Input('attribute-filter-{}'.format(attributes_column), 'value') for attributes_column in attributes_columns])
def render_content(variants, tab, start_date, end_date, contained, 
                   start_steps, end_steps, case_performance, 
                   excluded_activities, time_before_activity, self_transition,
                   *attributes):
    global current_process
    ctx = dash.callback_context
    if ctx.triggered[0]["prop_id"] == 'tabs.value':
        if tab != 'process-discovery':
            return None, None, variants_filter_object(tab=tab)
    
    
    filtered_workflow = apply_filters(workflow_df, start_date, end_date, contained, 
                                       start_steps, end_steps, case_performance, 
                                       excluded_activities, time_before_activity, self_transition,
                                       *attributes)
    
    filters = {'start_date': start_date, 'end_date': end_date,
              'contained': contained, 
               'start_steps': start_steps, 'end_steps': end_steps, 
               'case_performance': case_performance, 
               'excluded_activities': excluded_activities, 
               'time_before_activity': time_before_activity, 
               'self_transition': self_transition}
    
    for i, attribute in enumerate(attributes):
        filters[attributes_columns[i]] = attribute
        
    nb_variants = case_statistics.get_variant_statistics(filtered_workflow,
                                          parameters={case_statistics.Parameters.CASE_ID_KEY: "case",
                                                      case_statistics.Parameters.ACTIVITY_KEY: "activity",
                                                      case_statistics.Parameters.TIMESTAMP_KEY: "timestamp"})
    
    if variants is None:
        variants = 10
    
    current_process = filtered_workflow.to_json()
    return filters, variants_filter_object(variants, len(nb_variants))

@app.callback(Output('filtered-workflow', 'selected_node'),
              Output('filtered-workflow', 'selected_edge'),
              Input('card-close-button', 'n_clicks'))
def reset_selection(n_clicks):
    if n_clicks is None:
        raise PreventUpdate
    else:
        return None, None

@app.callback(Output('selected', 'children'),
              Output('selected', 'style'),
              Input('filtered-workflow', 'selected_node'),
              Input('filtered-workflow', 'selected_edge'),
              Input('card-close-button', 'n_clicks'),
              State('unique-nodes', 'data'))
def output_selected(selected_node, selected_edge, n_close, data):
    ctx = dash.callback_context

    if not ctx.triggered:
        raise PreventUpdate
    
    button = ctx.triggered[0]["prop_id"]
    if button == 'card-close-button.n_clicks':
        return [dbc.CardHeader(
                        [
                            "",
                            html.Button(
                                html.I(className="fas fa-times"),
                                className="ml-auto close",
                                id='card-close-button',
                            style={'float': 'right'}),
                        ]),
                        dbc.CardBody('')], {'position': 'absolute',
      'z-index': '999999',
      'display': 'None'}
    if selected_node is not None:
        node = data[int(selected_node)]
        return [dbc.CardHeader(
                dbc.Row([
                    dbc.Col(filter_title('Node: ' + node), md=10),
                    dbc.Col(html.Button(
                        html.I(className="fas fa-times"),
                        className="ml-auto close",
                        id='card-close-button',
                        style={'border-width': '0px',
                              'background-color': '#f7f7f7'}
                    ), md=2)
                ])),
                dbc.CardBody(selected_node_window(selected_node, data, activities_count_df, transitions_df, total_cases))], {'position': 'absolute',
                      'z-index': '999999',
                      'display': 'block',
                      'margin-top': '10px',
                      'margin-left': '10px'}
    elif selected_edge is not None:
        nodes = selected_edge.split('->')
        source = data[int(nodes[0])]
        target = data[int(nodes[1])]
        return [dbc.CardHeader(
                dbc.Row([
                    dbc.Col(filter_title('Transition: ' + source + ' -> ' + target),
                            md=10),
                    dbc.Col(html.Button(
                            html.I(className="fas fa-times"),
                            className="ml-auto close",
                            id='card-close-button',
                        style={'border-width': '0px',
                              'background-color': '#f7f7f7'}
                    ))
                ])),
                dbc.CardBody(selected_edge_window(nodes, data, transitions_df, total_cases))], {'position': 'absolute',
                      'z-index': '999999',
                      'display': 'block',
                      'margin-top': '10px',
                      'margin-left': '10px'}
    else:
        return [dbc.CardHeader(
                        [
                            "",
                            html.Button(
                                html.I(className="fas fa-times"),
                                className="ml-auto close",
                                id='card-close-button',
                            ),
                        ]),
                        dbc.CardBody('')], {'position': 'absolute',
      'z-index': '999999',
      'display': 'None'}

@app.callback(Output('selected-individual', 'children'),
              Output('selected-individual', 'style'),
              Input('individual-workflow', 'selected_node'),
              Input('individual-workflow', 'selected_edge'),
              Input('card-close-individual-button', 'n_clicks'),
              State('unique-nodes-individual', 'data'))
def output_selected(selected_node, selected_edge, n_close, data):
    ctx = dash.callback_context

    if not ctx.triggered:
        raise PreventUpdate
    
    button = ctx.triggered[0]["prop_id"]
    if button == 'card-close-individual-button.n_clicks':
        return [dbc.CardHeader(
                        [
                            "",
                            html.Button(
                                html.I(className="fas fa-times"),
                                className="ml-auto close",
                                id='card-close-individual-button',
                            style={'float': 'right'}),
                        ]),
                        dbc.CardBody('')], {'position': 'absolute',
      'z-index': '999999',
      'display': 'None'}
    if selected_node is not None:
        node = data[int(selected_node)]
        return [dbc.CardHeader(
                dbc.Row([
                    dbc.Col(filter_title('Node: ' + node), md=10),
                    dbc.Col(html.Button(
                        html.I(className="fas fa-times"),
                        className="ml-auto close",
                        id='card-close-individual-button',
                        style={'border-width': '0px',
                              'background-color': '#f7f7f7'}
                    ), md=2)
                ])),
                dbc.CardBody(selected_node_window(selected_node, data, activities_count_df, transitions_df, total_cases))], {'position': 'absolute',
                      'z-index': '999999',
                      'display': 'block',
                      'margin-top': '10px',
                      'margin-left': '10px'}
    elif selected_edge is not None:
        nodes = selected_edge.split('->')
        source = data[int(nodes[0])]
        target = data[int(nodes[1])]
        return [dbc.CardHeader(
                dbc.Row([
                    dbc.Col(filter_title('Transition: ' + source + ' -> ' + target),
                            md=10),
                    dbc.Col(html.Button(
                            html.I(className="fas fa-times"),
                            className="ml-auto close",
                            id='card-close-individual-button',
                        style={'border-width': '0px',
                              'background-color': '#f7f7f7'}
                    ))
                ])),
                dbc.CardBody(selected_edge_window(nodes, data, transitions_df, total_cases))], {'position': 'absolute',
                      'z-index': '999999',
                      'display': 'block',
                      'margin-top': '10px',
                      'margin-left': '10px'}
    else:
        return [dbc.CardHeader(
                        [
                            "",
                            html.Button(
                                html.I(className="fas fa-times"),
                                className="ml-auto close",
                                id='card-close-individual-button',
                            ),
                        ]),
                        dbc.CardBody('')], {'position': 'absolute',
      'z-index': '999999',
      'display': 'None'}


dropdown_excluded_activities = dcc.Dropdown(id='excluded-activities-dropdown',
                             options=[
           {'label': activity, 'value': activity}
        for activity in activity_count_df['activity']],
                             value=events_filters,
                             multi=True)

@app.callback(Output('loading', 'children'),
             Input('push-to-flow', 'n_clicks'),
             State('excluded-activities-dropdown', 'value'),
             State('time-before-activity-filter', 'value'),
             State('self-transition', 'on'))
def push_to_flow(n_clicks, excluded_activities, time_before, self_transition):
    if n_clicks==0:
        raise PreventUpdate
    else:
        variables['standard']['filtered_events'] = excluded_activities
        variables['standard']['min_time_filter'] = time_before
        variables['standard']['self_transition'] = self_transition
        project.set_variables(variables)
        scenario = project.get_scenario("Rebuild_Flow")
        scenario.run_and_wait()

@app.callback(Output('save-process', 'style'),
              State('save-process', 'style'),
              Input('saved-process-name', 'value')) 
def desactivate_button(style, value):
    if style is None:
        raise PreventUpdate
    if value is None:
        style['background-color'] = '#0d6efd'
        style['border-color'] = '#0d6efd'
    elif len(value) == 0:
        style['background-color'] = '#6c757d'
        style['border-color'] = '#6c757d'
    else:
        style['background-color'] = '#0d6efd'
        style['border-color'] = '#0d6efd'
    return style

def saved_processes_dropdown(add_value=False, default_value=None):
    input_folder = dataiku.Folder("1rhVX912")
    paths = input_folder.list_paths_in_partition()
    names = [name.replace('_petri.p', '').replace('/', '') for name in paths if '_petri.p' in name]
    if default_value is None and len(names)>0:
        default_value = names[0]
    if add_value:
        names.append('Save current process')
    saved_processes_dropdown_object = dcc.Dropdown(
            id='saved-processes-dropdown',
            options=[{'label': name, 'value': name} for name in names],
            value=default_value)
    return saved_processes_dropdown_object

dropdown_menu = dbc.DropdownMenu(
                        id='process-menu',
                        label="Action",
                        children=[
                            dbc.DropdownMenuItem("Apply", id='apply-button'),
                            dbc.DropdownMenuItem("Delete", id='delete-button'),
                        ],
                        size='sm'
                    )

dropdown_menu_conformance = dbc.DropdownMenu(
                        id='process-menu',
                        label="Action",
                        children=[
                            dbc.DropdownMenuItem("Apply", id='apply-button'),
                            dbc.DropdownMenuItem("Run Conformance Checks", id='conformance-button'),
                            dbc.DropdownMenuItem("Delete", id='delete-button'),
                        ],
                        size='sm'
                    )

def process_dropdown(add_value=True, default_value=None, tab='process-discovery'):
    if tab == 'process-discovery':
        title = 'Reference process'
        dropdown = dropdown_menu
    elif tab == 'conformance-checks':
        title = 'Compare reference process'
        dropdown = dropdown_menu_conformance
    else:
        title = 'Compare reference process'
        dropdown = dropdown_menu
    return dbc.Col([html.P(title, style={'font-weight': 'bold'}),
            dbc.Row([dbc.Col(id='saved-processes-container',
                             children=saved_processes_dropdown(add_value, default_value), 
                             style={'padding': 0}, 
                             md=8),
                    dbc.Col(id='process-button', 
                            children=dropdown, md=4)], 
                   style={'margin-top': '7px', 'margin-bottom': '7px',
                         'margin-left': 0, 'margin-right': 0})
                   ])

@app.callback(Output('process-button', 'children'),
              Output('saved-processes-value', 'data'),
              Input('tabs', 'value'),
              Input('saved-processes-dropdown', 'value'),
              State('saved-processes-value', 'data'))
def update_button(tab, process, old_process):
    ctx = dash.callback_context
    if ctx.triggered[0]['prop_id'] == 'saved-processes-dropdown.value':
        if process == old_process:
            raise PreventUpdate
    if process == 'Save current process' and tab == 'process-discovery':
        return [dbc.Row(
            [dbc.Col(process_input_name, md=8, 
                     style={'padding-left': '6px', 'padding-right': '6px'}), 
                dbc.Col(dcc.Loading(id='saving', type='default', 
                           children=[
             dbc.Button("Save", 
                        id='save-process', 
                        n_clicks=0,
                        size='sm')]), md=4,
                        style={'padding-left': 0, 'padding-right': '6px'})]),
               html.Div(dropdown_menu, style={'display': 'None'})], process
    elif tab != 'conformance-checks':
        return [dbc.Button("Save", 
                        id='save-process', 
                        n_clicks=0,
                        size='sm',
                        style={'display': 'None'}),
               html.Div(process_input_name, style={'display': 'None'}),
               dropdown_menu], process
    else:
        return [dbc.Button("Save", 
                        id='save-process', 
                        n_clicks=0,
                        size='sm',
                        style={'display': 'None'}),
                html.Div(process_input_name, style={'display': 'None'}),
                dropdown_menu_conformance,
               ], process

process_mining_explanation = html.P('Use filters to discover your process through the data. Remove noisy activities through the bottom filters. Processes displayed in the main screen can be saved as reference and compared through conformance checks and individual traces.')

@app.callback(Output('process-header', 'children'),
              Input('tabs', 'value'),
              State('saved-processes-value', 'data'),
              State('apply-button', 'n_clicks'))
def apply_or_update_process(tab, process, n_clicks):
    ctx = dash.callback_context
    if not ctx.triggered:
        raise PreventUpdate
    if ctx.triggered[0]['prop_id'] == 'apply-button.n_clicks':
        if process == 'Save current process':
            raise PreventUpdate
    if tab == 'process-discovery':
        return process_dropdown(True, process)
    else:
        return process_dropdown(False, process, tab)
    
    
@app.callback(Output('base-filters', 'children'),
              State('tabs', 'value'),
              State('saved-processes-value', 'data'),
              State('base-filters', 'children'),
              Input('apply-button', 'n_clicks'))
def apply_filters_button(tab, process, old_base_filters, n_clicks):
    ctx = dash.callback_context
    if n_clicks is None:
        raise PreventUpdate
    filters_data = read_pickle_from_dss_folder(process + '_filters.p', '1rhVX912')
    base_filters = [filter_element('Starting at', single_datetime_filter_picker(filters_data['start_date'], True)),
                     filter_element('Ending at', single_datetime_filter_picker(filters_data['end_date'], False)),
                     filter_element('Strictly contained', contained_intersecting(filters_data['contained'])),
                     filter_element('Starting activities', start_step(start_end_df, filters_data['start_steps'])),
                     filter_element('Ending activities', end_step(start_end_df, filters_data['end_steps'])),
                     filter_element('Traces duration', case_performance_slider(filters_data['case_performance'])),
                     html.Div(attribute_filters({attribute: filters_data[attribute] for attribute in attributes_columns}), id='attribute-filters')]
    return base_filters

@app.callback(Output('min-traces-duration', 'children'),
              Output('max-traces-duration', 'children'),
              Input('case-performance-filter', 'value'))
def output_duration_bounds(case_performance):
    return html.P(formatted_human_readable(case_performance_scale[case_performance[0]])), html.P(formatted_human_readable(case_performance_scale[case_performance[1]]))

@app.callback([Output('min-attribute-filter-{}'.format(attributes_column), 'children') for attributes_column in attributes_columns if is_numeric_dtype(workflow_df[attributes_column])],
              [Output('max-attribute-filter-{}'.format(attributes_column), 'children') for attributes_column in attributes_columns if is_numeric_dtype(workflow_df[attributes_column])],
              Output('dummy-output', 'data'),
              Input('dummy-input', 'data'),
              [Input('attribute-filter-{}'.format(attributes_column), 'value') for attributes_column in attributes_columns if is_numeric_dtype(workflow_df[attributes_column])])
def output_duration_bounds(dummy_input, values):
    if len(attribute_scales) == 0:
        return [dummy_input]
    if len(attribute_scales) == 1:
        values = [values]
    columns = [attributes_column for attributes_column in attributes_columns if is_numeric_dtype(workflow_df[attributes_column])]
    return [html.P(human_format(attribute_scales[columns[i]][value[0]])) for i, value in enumerate(values)], [html.P(human_format(attribute_scales[columns[i]][value[1]])) for i, value in enumerate(values)], dummy_input

def filter_header():
    return html.Div(children=process_mining_explanation, 
                    id='filter-header')

def process_header_menu():
    return html.Div(children=process_dropdown(),
                       id='process-header')

def variants_container():
    return html.Div(children=variants_filter_object(),
                              id='variants-container')

def filter_subtitle():
    return html.Div(children=html.P('Select traces', style={'font-weight': 'bold', 'clear': 'both', 'padding-top': '12px'}),
                           id='filter-subtitle')

base_filters = html.Div(id='base-filters',
                        children=[
                         filter_element('Starting at', 
                                        single_datetime_filter_picker(timeframe_df['timestamp_min'].iloc[0], True)),
                         filter_element('Ending at', 
                                        single_datetime_filter_picker(timeframe_df['timestamp_max'].iloc[0], False)),
                         filter_element('Strictly contained', 
                                        contained_intersecting(True)),
                         filter_element('Starting activities', start_step(start_end_df, None)),
                         filter_element('Ending activities', end_step(start_end_df, None)),
                         filter_element('Traces duration', case_performance_slider([0, 100])),
                         html.Div(attribute_filters(), id='attribute-filters')])

additional_filters = html.Div(id='additional-filters',
    children=[
            html.Hr(),
            html.P('Excluded activities', style={'font-weight': 'bold'}),
            html.P('Filter out unwanted noisy activities'),
            filter_element('Exclude activities',
            html.Div(id='excluded-activities', children=[dropdown_excluded_activities])),
            time_before_filter_object(min_time_filter),
            html.Div(style={'clear': 'both'}),
            filter_element('Looping activities',
                           remove_self_transition),
            html.P('Filter out activities that occur after the same', style={'color': '#BBBBBB', 'font-size': '12px'}),
            dbc.Button("Remove excluded activities from dataset", 
                        id='push-to-flow', 
                        n_clicks=0,
                        size='sm',
                        style=button_style_bottom),
             html.P('Refresh the page once pushed', 
                    style={'color': '#bbbbbb', 'font-size': '12px', 
                           'text-align': 'center', 'margin-bottom': 0}),
             html.Hr()])

conformance_check_filters = dbc.Col(base_filters, 
                                   style={'margin-left': 0, 
                                          'border-right': '1px solid',
                                          'border-right-color': '#e3e4e4'})

conformance_granularity = dcc.RadioItems(options=[{'label': 'Monthly', 'value': 'monthly'},
                                                  {'label': 'Weekly', 'value': 'weekly'},
                                                  {'label': 'Daily', 'value': 'daily'}],
                                         value='monthly',
                                         id='conformance-granularity',
                                        inputStyle={"margin": "7px"})

@app.callback(Output('individual-processes', 'children'),
              Input('saved-processes-value', 'data'),
              Input('start-date-filter', 'value'),
              Input('end-date-filter', 'value'),
              Input('contained-intersecting', 'value'),
              Input('start-step', 'value'),
              Input('end-step', 'value'),
              Input('case-performance-filter', 'value'),
              [Input('attribute-filter-{}'.format(attributes_column), 'value') for attributes_column in attributes_columns])
def individual_processes_table(process, start_date, end_date, contained, 
                               start_steps, end_steps, case_performance, 
                               *attributes):
    global individual_data
    filtered_workflow = apply_filters(workflow_df, start_date, end_date, contained, 
                                               start_steps, end_steps, case_performance, 
                                               events_filters, min_time_filter, self_transition,
                                               *attributes)
    case_ids = [str(x) for x in filtered_workflow['case'].unique()]
    conformance_checks = dataiku.Dataset('process_conformance_checks')
    try:
        conformance_checks_df = conformance_checks.get_dataframe()
    except:
        return html.P('Conformance checks have not been computed for ' + process + ' process. ' +
                      'Please run conformance checks in the Action menu.')
    if process not in conformance_checks_df['reference_process'].unique():
        return html.P('Conformance checks have not been computed for ' + process + ' process. ' +
                      'Please run conformance checks in the Action menu.')
    checks = conformance_checks_df[conformance_checks_df['case'].isin(case_ids)]
    checks = checks[checks['reference_process']==process]
    table_columns = ['case', 'trace_fitness', 'timestamp'] + attributes_columns
    table_columns_names = ['Case', 'Fitness', 'Start Date'] + attributes_columns
    individual_data = [{col: checks[col].iloc[i] for col in table_columns}
                  for i in range(len(checks))]
    individual_processes_table = dash_table.DataTable(
            id='individual-processes-table',
            columns=(
                    [{'id': table_columns[i], 'name': table_columns_names[i]}
                     for i in range(len(table_columns))]
                 ),
            data=individual_data,
        style_cell={'textAlign': 'left',
                    'font-family': 'Helvetica Neue'},
        active_cell={'column': 0, 'row': 0},
        sort_action='native',
        style_table={'height': '20em', 'overflowY': 'auto', 'cursor': 'pointer'},
        style_data_conditional=[
            {
                'if': {
                    'state': 'active'
                },
                'backgroundColor': '#eef5ff',
                "border": "1px solid #288cfe"
            },
            {
                'if': {
                    'state': 'selected'
                },
                'backgroundColor': '#eef5ff',
                "border": "1px solid #288cfe"
            },
        ]
    )
    explanation = html.P('The table displays the traces filtered through the left panel filter. ' +
                         'The fitness score is computed with respect to the selected process. ' +
                         'Individual processes are visualized below the table. ' +
                         'The Fitness column is a number between 0 and 1 assessing the fit of a given trace with the ' + process + ' process. ' +
                         'Click on a row to display process.',
                        style={'padding': '32px', 'padding-bottom': 0})
    return [explanation, individual_processes_table]

@app.callback(Output('conformance-explanation', 'children'),
              Input('saved-processes-value', 'data'))
def update_explanation(process):
    return html.P('Traces filtered with filters in left panel are checked against the selected reference process. ' +
                  'Overall conformance is plotted in the main screen. ' + 
                  'The graph represents the overall conformance of the traces defined by the left-hand side filters ' + 
                  'with respect to the ' + process + ' process.')

@app.callback(Output('graphs', 'children'),
              Input('tabs', 'value'),
              Input('conformance-button', 'n_clicks'),
              State('saved-processes-value', 'data'),
              Input('conformance-granularity', 'value'),
              Input('start-date-filter', 'value'),
              Input('end-date-filter', 'value'),
              Input('contained-intersecting', 'value'),
              Input('start-step', 'value'),
              Input('end-step', 'value'),
              Input('case-performance-filter', 'value'),
              [Input('attribute-filter-{}'.format(attributes_column), 'value') for attributes_column in attributes_columns])
def make_figures(tab, n_clicks, process, conformance_granularity,
                 start_date, end_date, contained, 
                 start_steps, end_steps, case_performance, 
                 *attributes):
    ctx = dash.callback_context
    if ctx.triggered[0]['prop_id'] == 'conformance-button.n_clicks' and n_clicks is not None:
        if n_clicks > 0:
            scenario = project.get_scenario("RUNCONFORMANCECHECKS")
            scenario.run_and_wait()
    conformance_checks = dataiku.Dataset("process_conformance_checks")
    try:
        conformance_checks_df = conformance_checks.get_dataframe()
    except:
        return html.P('Conformance checks have not been computed for ' + process + ' process. ' +
                      'Please run conformance checks in the Action menu.')
    if process not in conformance_checks_df['reference_process'].unique():
        return html.P('Conformance checks have not been computed for ' + process + ' process. ' +
                      'Please run conformance checks in the Action menu.')
    filtered_workflow = apply_filters(workflow_df, start_date, end_date, contained, 
                                           start_steps, end_steps, case_performance, 
                                           events_filters, min_time_filter, self_transition,
                                           *attributes)
    case_ids = filtered_workflow['case'].unique()
    checks = conformance_checks_df[conformance_checks_df['reference_process']==process]
    checks = checks[checks['case'].isin(case_ids)]
    if conformance_granularity == 'monthly':
        checks['timestamp'] = [datetime(d.year, d.month, 1) for d in checks['timestamp']]
        fitness_date = checks.groupby(['timestamp'])['trace_fitness'].mean().reset_index()
    elif conformance_granularity == 'weekly':
        checks['timestamp'] = [start_week(d) for d in checks['timestamp']]
        fitness_date = checks.groupby(['timestamp'])['trace_fitness'].mean().reset_index()
    else:
        checks['timestamp'] = [datetime(d.year, d.month, d.day) for d in checks['timestamp']]
        fitness_date = checks.groupby(['timestamp'])['trace_fitness'].mean().reset_index()
    global_fitness = checks['trace_fitness'].mean()
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=fitness_date["timestamp"], y=fitness_date["trace_fitness"], 
                  name='Fitness Evolution (Average Fitness ' + str(round(global_fitness, 3)) + ')',
                  line=dict(color=palette[0])))
    fig.update_layout(plot_bgcolor='rgba(0,0,0,0)')
    fig.layout.yaxis.gridcolor = '#D7DBDE'
    fig.layout.margin.t = 10
    return dbc.Col(dcc.Graph(figure=fig, 
                             id='conformance-checks-evolution',
                            style={'margin': 0}))

@app.callback(Output('saved-processes-container', 'children'),
              Input('save-process', 'n_clicks'),
              Input('delete-button', 'n_clicks'),
              State('saved-process-name', 'value'),
              State('filters', 'data'),
              Input('tabs', 'value'),
              State('saved-processes-value', 'data'),
              State('saved-processes-dropdown', 'options'))
def update_process(save_button, delete_button, name, filters, tab, process, options):
    ctx = dash.callback_context
    option_labels = [option['label'] for option in options]
    if 'Save current process' in option_labels and tab != 'process-discovery':
        return saved_processes_dropdown(False, default_value=process)
    elif 'Save current process' not in option_labels and tab == 'process-discovery':
        return saved_processes_dropdown(True, default_value=process)
    elif ctx.triggered[0]['prop_id'] == 'save-process.n_clicks':
        if save_button is None or name is None:
            raise PreventUpdate
        elif save_button >= 0 and len(name) > 0:
            data = pd.read_json(final_process)
            event_log = pm4py.format_dataframe(data, case_id='case', activity_key='activity', timestamp_key='timestamp')
            net, im, fm = inductive_miner.apply(event_log, variant=inductive_miner.Variants.IM_CLEAN)
            workflow_mined, start_end, activities = mine_process(data, 'case', 'activity', 'timestamp', transition='performance')
            folder_id = get_managed_folder_id_with_folder_name(project, "saved_processes")
            petri_net = {'pn': net, 'im': im, 'fm': fm}
            write_pickle_in_dss_folder(petri_net, name + '_petri.p', folder_id)
            dfg = {'workflow_mined': workflow_mined, 'start_end': start_end, 'activities': activities}
            write_pickle_in_dss_folder(dfg, name + '_dfg.p', folder_id)
            write_pickle_in_dss_folder(filters, name + '_filters.p', folder_id)
            return saved_processes_dropdown(True, default_value=name)
    elif ctx.triggered[0]['prop_id'] == 'delete-button.n_clicks':
        if delete_button is None:
            raise PreventUpdate
        elif delete_button == 0:
            raise PreventUpdate
        elif delete_button > 0:
            input_folder = dataiku.Folder("1rhVX912")
            input_folder.delete_path(process + '_dfg.p')
            input_folder.delete_path(process + '_filters.p')
            input_folder.delete_path(process + '_petri.p')
            return saved_processes_dropdown(True, default_value=[option['label'] for option in options if option['label'] != 'process'][0])
    else:
        raise PreventUpdate

@app.callback(Output('filter-subtitle', 'children'),
              Input('tabs', 'value'))
def update_title(tab):
    if tab == 'process-discovery':
        title = 'Select traces'
    elif tab == 'conformance-checks':
        title = 'to traces'
    else:
        title = 'to individual traces'
    return html.P(title, style={'font-weight': 'bold', 'clear': 'both', 'padding-top': '12px'})

@app.callback(Output('filter-header', 'children'),
              Input('tabs', 'value'))
def update_title(tab):
    if tab == 'process-discovery':
        return process_mining_explanation
    else:
        return None

def layout_discovery():
    return [dbc.Row([dbc.Col(id='nb-cases-info', md=4),
                     dbc.Col(id='graph-legend', 
                             children=legend_object(1, 100),
                             md=6),
                     dbc.Col(dbc.Row([
                         dbc.Col(html.P('Frequency', style={'float': 'right',
                                                           'padding': 0}), style={'padding': 0}, md=3),
                         dbc.Col(daq.BooleanSwitch(id='frequency-time',
                              on=False), style={'padding': 0}, md=6),
                         dbc.Col(html.P('Time', style={'padding': 0}), 
                                 style={'padding': 0}, md=3)]), md=2)], 
                    style={'margin-top': '7px'}),
            dbc.Row([dbc.Col([html.Div(style={'height': '16px', 'width': '16px', 
                                             'background-color': '#767676', 'transform': 'rotate(45deg)', 
                                              'float': 'left', 'margin-left': '5px'}),
                             html.P('Starting activities', style={'margin-left': '10px', 'float': 'left'})], md=3),
                    dbc.Col([html.Div(style={'height': '16px', 'width': '16px', 
                                             'background-color': '#767676', 
                                              'float': 'left', 'margin-left': '5px'}),
                             html.P('Activities', style={'margin-left': '10px', 'float': 'left'})], md=3),
                    dbc.Col([html.Div(style={'width': 0,
                                            'height': 0,
                                            'border-left': '8px solid transparent',
                                            'border-right': '8px solid transparent',
                                            'border-bottom': '16px solid #767676',
                                              'float': 'left', 'margin-left': '5px'}),
                             html.P('Ending activities', style={'margin-left': '10px', 'float': 'left'})], md=3),
                    ]),
            html.Hr(),
            html.P('You can click on activities and transitions to see details', style={'color': '#bbb'}),
             dbc.Card(id='selected',
                      className='w-25',
                      children=[dbc.CardHeader(
                        [
                            "",
                            html.Button(
                                className="ml-auto close",
                                id='card-close-button',
                            ),
                        ]),
                        dbc.CardBody('')],
                     style={'position': 'absolute',
                            'z-index': '999999',
                            'display': 'None'}),
             dcc.Loading(id='graph-loading', type='circle', 
                         children=html.Div(id='graph-workflow'))]
    
def layout_conformance():
    return [dbc.Container(id='conformance-explanation'), conformance_granularity, dcc.Loading(id='graphs', type='circle')]

@app.callback(Output('individual-graph', 'children'),
              Output('graph-legend-individual', 'children'),
              Output('unique-nodes-individual', 'data'),
              State('individual-processes-table', 'derived_virtual_indices'),
              Input('individual-processes-table', 'active_cell'),
              Input('frequency-time-individual', 'on'))
def render_graph(indices, cell, frequency_time):
    if indices is not None and cell is not None:
        case = individual_data[indices[cell['row']]]['case']
        filtered_workflow = workflow_df[workflow_df['case']==case]
        workflow_mined, start_end, activities = mine_process(filtered_workflow, 'case', 'activity', 'timestamp', transition=('performance' if frequency_time else 'frequency'))
        graph, unique_nodes, max_count = create_graph(workflow_mined, start_end, activities, transition=('time' if frequency_time else 'frequency'))
        graph_object = dash_interactive_graphviz.DashInteractiveGraphviz(
            id="individual-workflow",
            dot_source=graph,
            style={'width': '67%', 'height': '67%',
                  'cursor': 'pointer'}
        )
        return graph_object, legend_object(1, max_count), unique_nodes

def layout_individual():
    return [dcc.Loading(id='individual-processes', type='circle'),
            dbc.Row([dbc.Col(id='empty-cell', md=4),
                 dbc.Col(id='graph-legend-individual', 
                         md=6),
                 dbc.Col(dbc.Row([
                     dbc.Col(html.P('Frequency', style={'float': 'right',
                                                       'padding': 0}), style={'padding': 0}, md=3),
                     dbc.Col(daq.BooleanSwitch(id='frequency-time-individual',
                          on=True), style={'padding': 0}, md=6),
                     dbc.Col(html.P('Time', style={'padding': 0}), 
                             style={'padding': 0}, md=3)]), md=2)], 
                style={'margin-top': '7px'}),
            html.Div(id='individual-graph'),
            dbc.Card(id='selected-individual',
                      className='w-25',
                      children=[dbc.CardHeader(
                        [
                            "",
                            html.Button(
                                className="ml-auto close",
                                id='card-close-individual-button',
                            ),
                        ]),
                        dbc.CardBody('')],
                     style={'position': 'absolute',
                            'z-index': '999999',
                            'display': 'None'})]

@app.callback(Output('tab-content', 'children'),
              Input('tabs', 'value'),
              State('saved-processes-value', 'data'))
def render_content(tab, process):
    if tab == 'process-discovery':
        return layout_discovery()
    elif tab == 'conformance-checks':
        if process is None:       
            return html.P('Save a process in the Process Discovery tab before performing some conformance checks.')
        else:
            return html.Div(layout_conformance(), 
                        style={'padding': '32px'})
    elif tab == 'individual-traces':
        if process is None:       
            return html.P('Save a process in the Process Discovery tab before performing some conformance checks.')
        else:
            return layout_individual()

@app.callback(Output('explanation', 'children'),
              Input('tabs', 'value'))
def render_content(tab):
    if tab == 'process-discovery':
        return html.H5('Process Mining',
                     style={'margin': 0, 'margin-top': '12px'})
    elif tab == 'conformance-checks':
        return html.H5('Process Mining',
                     style={'margin': 0, 'margin-top': '12px'})
    elif tab == 'individual-traces':
        return html.H5('Process Mining',
                     style={'margin': 0, 'margin-top': '12px'})

@app.callback(Output('additional-filters', 'children'),
              Input('tabs', 'value'))
def render_content(tab):
    if tab == 'process-discovery':
        return additional_filters
    else:
        return None

layout_mining = html.Div([
    dcc.Loading(id='loading', type='cube', fullscreen=True),
    dbc.Row([dbc.Col(id='explanation', children=[html.P('')], 
                     style={'padding': '24px'}, md=5),
             dbc.Col(dcc.Tabs(id="tabs", 
                             value='process-discovery', children=[
                    dcc.Tab(label='Process Discovery', 
                            value='process-discovery'),
                    dcc.Tab(label='Conformance Checks', 
                            value='conformance-checks'),
                    dcc.Tab(label='Individual Traces', 
                            value='individual-traces')],
                             style={'border-bottom': 'none',
                                    'padding': 0}), style={'padding': 0, 'padding-top': '24px'}, md=7)]),
    dbc.Row([dbc.Col(html.Hr(style={'padding': 0, 'margin': 0}), style={'padding': 0}, md=5),
             dbc.Col(md=7)]),
    dbc.Row([dbc.Col(children=[
                     filter_header(),
                     process_header_menu(),
                     variants_container(),
                     filter_subtitle(),
                     base_filters, 
                     additional_filters], 
             style={'padding': '24px', 
                   'border-right': '1px solid',
                   'border-right-color': '#e3e4e4'},
                     md=4),
        dbc.Col(id='tab-content', 
                md=8)]),
    dcc.Store(id='filters'),
    dcc.Store(id='unique-nodes'),
    dcc.Store(id='unique-nodes-individual'),
    dcc.Store(id='old-variants-filter', data=10),
    dcc.Store(id='saved-processes-value'),
    dcc.Store(id='dummy-input', data=1),
    dcc.Store(id='dummy-output', data=1)],
             style={'font-size': '12px', 'font-family': 'Helvetica Neue'})

# build your Dash app
app.layout = layout_mining